/*
 * Xlib interface for QEMU System Emulator
 *
 * Copyright (c) 2004 Vladimit Oleynik <dzo@simtreas.ru>
 * Used ideas from Bochs, SDL, PCEmu.
*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

/* hack, special for qemu */
#define X_display_init sdl_display_init

#include "vl.h"

static int last_vm_running;

static Display *qx_display;
static int q_x_screen_num;
static Window win;
static GC gc;
static XImage *ximage = NULL;
static int     default_depth;
static Visual  *default_visual;
static int width, height;

static void x_update(DisplayState *ds, int x, int y, int w, int h)
{
    XPutImage(qx_display, win, gc, ximage, x, y, x, y, w, h);
    XFlush(qx_display);
}

static int prev_x, prev_y;
static int current_x, current_y;

/* Move the cursor to a center window place */
static void qXWarpPointer(void)
{
	int x = width/2;
	int y = height/2;

	prev_x = current_x = x;
	prev_y = current_y = y;
	XWarpPointer(qx_display, None, win, 0, 0, 0, 0, x, y);
}

static void x_resize(DisplayState *ds, int w, int h)
{
    XSizeHints hints;
    long supplied_return;

    if ( XGetWMNormalHints(qx_display, win, &hints, &supplied_return) &&
	 supplied_return & PMaxSize ) {
      hints.max_width = hints.min_width = w;
      hints.max_height = hints.min_height = h;
      XSetWMNormalHints(qx_display, win, &hints);
    }
    if(ximage) {
	XDestroyImage(ximage);
//        free(ds->data);
	XResizeWindow(qx_display, win, w, h);
    }
    ximage = XCreateImage(qx_display, default_visual,
	     default_depth,          // depth of image (bitplanes)
	     ZPixmap,
	     0,                      // offset
	     NULL,                   // malloc() space after
	     w, h,                   // x & y size of image
	     32,                     // # bits of padding
	     0 );                    // bytes_per_line, let X11 calculate
    if (!ximage) {
	fprintf(stderr, "Could not XCreateImage()\n");
	exit(1);
    }
    width = ds->width = w;
    height = ds->height = h;
    ds->depth = ximage->bits_per_pixel;
    ds->linesize = ximage->bytes_per_line;
    ds->data = (char *) malloc( (size_t) (ximage->bytes_per_line * h) );
    if (!ds->data) {
	fprintf(stderr, "Could not malloc Image data\n");
	exit(1);
    }
    ximage->data = ds->data;
    qXWarpPointer();
    XFlush(qx_display);
}

static int gui_grab;

static void x_SetCaption(void)
{
	char *title, *icon;
	XTextProperty titleprop, iconprop;

	title = icon = "QEMU";
	if (!vm_running && !gui_grab)
		title = "QEMU [Stopped]";
	if (vm_running && gui_grab)
		title = "QEMU - Press Ctrl-Shift to exit grab";
	if (!vm_running && gui_grab)
		title = "QEMU [Stopped] - Press Ctrl-Shift to exit grab";

	XStringListToTextProperty(&title, 1, &titleprop);
	XSetWMName(qx_display, win, &titleprop);
	XFree(titleprop.value);

	XStringListToTextProperty(&icon, 1, &iconprop);
	XSetWMIconName(qx_display, win, &iconprop);
	XFree(iconprop.value);
	XSync(qx_display, /* no discard */ 0);
}

static void trigger_mouse_grab(void)
{
  gui_grab ^= 1;
  if (gui_grab) {
      /* disable cursor */
      static Cursor cursor;
      static int cursor_created;

#define shape_width 16
#define shape_height 16
#define mask_width 16
#define mask_height 16

      static unsigned int shape_bits[(16*16)/32];
      static unsigned int mask_bits[(16*16)/32];

      if (!cursor_created) {
	    Pixmap shape, mask;
	    XColor white, black;
	    Colormap default_cmap;

	    default_cmap = DefaultColormap(qx_display, q_x_screen_num);

	    shape = XCreatePixmapFromBitmapData(qx_display,
						RootWindow(qx_display,q_x_screen_num),
						(char*)shape_bits,
						shape_width, shape_height,
						1, 0, 1);
	    mask =  XCreatePixmapFromBitmapData(qx_display,
						RootWindow(qx_display,q_x_screen_num),
						(char*)mask_bits,
						mask_width, mask_height,
						1, 0, 1);
	    XParseColor(qx_display, default_cmap, "black", &black);
	    XParseColor(qx_display, default_cmap, "white", &white);
	    cursor = XCreatePixmapCursor(qx_display, shape, mask,
					 &white, &black, 1, 1);
	    cursor_created = 1;
      }
      XDefineCursor(qx_display, win, cursor);

      /* grab pointer and keyboard */

//    if ( flags & FULLSCREEN ) {
//      /* Unbind the mouse from the fullscreen window */
//      XUngrabPointer(qx_display, CurrentTime);
//    }
      /* Try to grab the mouse */
      while ( 1 ) {
	int result = XGrabPointer(qx_display, win, True, 0,
					GrabModeAsync, GrabModeAsync,
					win, None, CurrentTime);
	if (result == GrabSuccess)
		break;
      }
      /* Now grab the keyboard */
      XGrabKeyboard(qx_display, win, True,
			GrabModeAsync, GrabModeAsync, CurrentTime);

      /* Raise the window if we grab the mouse */
//    if ( !(flags & FULLSCREEN) )
	XRaiseWindow(qx_display, win);
      qXWarpPointer();
  } else {
    /* enable cursor */
    XUndefineCursor(qx_display, win);
    /* ungrab pointer and keyboard */
    XUngrabPointer(qx_display, CurrentTime);
    XUngrabKeyboard(qx_display, CurrentTime);
  }
  x_SetCaption();
}

static unsigned mouse_button_state;

static void send_keyboard_mouse_status(void)
{
    int dx, dy;

    dx = current_x - prev_x;
    dy = -(current_y - prev_y);
    kbd_mouse_event(dx, -dy, 0, mouse_button_state);
    if(current_x <= 1 || current_x >= (width-1) || current_y <= 1 || current_y >= (height-1)) {
	qXWarpPointer();
    } else {
	prev_x = current_x;
	prev_y = current_y;
    }
}

static const unsigned char scan_table1[] =
{
    0x39, 0x02,
#ifdef KBUK             /* double quotes, hash symbol */
    0x03, 0x2b,
#else
    0x28, 0x04,
#endif
    0x05, 0x06, 0x08, 0x28,
    0x0a, 0x0b, 0x09, 0x0d, 0x33, 0x0c, 0x34, 0x35,
    0x0b, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
    0x09, 0x0a, 0x27, 0x27, 0x33, 0x0d, 0x34, 0x35,
#ifdef KBUK             /* at symbol */
    0x28,
#else
    0x03,
#endif
    0x1e, 0x30, 0x2e, 0x20, 0x12, 0x21, 0x22,
    0x23, 0x17, 0x24, 0x25, 0x26, 0x32, 0x31, 0x18,
    0x19, 0x10, 0x13, 0x1f, 0x14, 0x16, 0x2f, 0x11,
    0x2d, 0x15, 0x2c, 0x1a,
#ifdef KBUK             /* backslash */
    0x56,
#else
    0x2b,
#endif
    0x1b, 0x07, 0x0c,
    0x29, 0x1e, 0x30, 0x2e, 0x20, 0x12, 0x21, 0x22,
    0x23, 0x17, 0x24, 0x25, 0x26, 0x32, 0x31, 0x18,
    0x19, 0x10, 0x13, 0x1f, 0x14, 0x16, 0x2f, 0x11,
    0x2d, 0x15, 0x2c, 0x1a,
#ifdef KBUK             /* vertical bar */
    0x56,
#else
    0x2b,
#endif

    0x1b,

#ifdef KBUK             /* tilde */
    0x2b,
#else
    0x29,
#endif
};


static const struct {
    KeySym key;
    unsigned scan_code;
} scan_table2[] = {
    { XK_BackSpace,     0x0e },
    { XK_Tab,           0x0f },
    { XK_Return,        0x1c },
    { XK_Escape,        0x01 },
    { XK_Delete,        0x53e0 },

    { XK_Home,          0x47e0 },
    { XK_Left,          0x4be0 },
    { XK_Up,            0x48e0 },
    { XK_Right,         0x4de0 },
    { XK_Down,          0x50e0 },
    { XK_Prior,         0x49e0 },
    { XK_Next,          0x51e0 },
    { XK_End,           0x4fe0 },
    { XK_Insert,        0x52e0 },
    { XK_Num_Lock,      0x45 },

/* This block of codes is for Sun Type 4 keyboards... */

    { XK_F27,           0x47e0 },       /* Home/R7/F27 */
    { XK_F29,           0x49e0 },       /* Prior/R9/F29 */
    { XK_F35,           0x51e0 },       /* Next/R15/F35 */
    { XK_F33,           0x4fe0 },       /* End/R13/F33 */

    { XK_F25,           0x36e0 },       /* Keypad divide/R5/F25 */
    { XK_F26,           0x37 },         /* Keypad multiply/R6/F26 */

    { XK_F23,           0x46 },         /* Scroll lock/R3/F23 */

    { XK_F31,           0x4c },         /* Keypad 5/R11/F31 */

/* End of Sun type 4 codes */

    { XK_KP_Enter,      0x1ce0 },
    { XK_KP_Multiply,   0x37 },
    { XK_KP_Divide,     0x35e0 },
    { XK_KP_Home,       0x47 },
    { XK_KP_Up,         0x48 },
    { XK_KP_Page_Up,    0x49 },
    { XK_KP_Subtract,   0x4a },
    { XK_KP_Left,       0x4b },
    { XK_KP_Begin,      0x4c },
    { XK_KP_Right,      0x4d },
    { XK_KP_Add,        0x4e },
    { XK_KP_End,        0x4f },
    { XK_KP_Down,       0x50 },
    { XK_KP_Page_Down,  0x51 },
    { XK_KP_Insert,     0x52 },
    { XK_KP_Decimal,    0x53 },
    { XK_KP_Delete,     0x53 },

    { XK_KP_0,          0x52 },
    { XK_KP_1,          0x4f },
    { XK_KP_2,          0x50 },
    { XK_KP_3,          0x51 },
    { XK_KP_4,          0x4b },
    { XK_KP_5,          0x4c },
    { XK_KP_6,          0x4d },
    { XK_KP_7,          0x47 },
    { XK_KP_8,          0x48 },
    { XK_KP_9,          0x49 },

    { XK_F1,            0x3b },
    { XK_Help,          0x3b }, /* openwin mapped */
    { XK_F2,            0x3c },
    { XK_F3,            0x3d },
    { XK_F4,            0x3e },
    { XK_F5,            0x3f },
    { XK_F19,           0x3c }, /* openwin mapped */
    { XK_F20,           0x3d }, /* openwin mapped */
    { XK_F16,           0x3e }, /* openwin mapped */
    { XK_F18,           0x3f }, /* openwin mapped */
    { XK_F6,            0x40 },
    { XK_F7,            0x41 },
    { XK_F8,            0x42 },
    { XK_F9,            0x43 },
    { XK_F10,           0x44 },
    { XK_F11,           0x57 },
    { XK_F12,           0x58 },

    { XK_Shift_L,       0x2a },
    { XK_Shift_R,       0x36 },
    { XK_Control_L,     0x1d },
    { XK_Control_R,     0x1de0 },
    { XK_Meta_L,        0x38 },
    { XK_Alt_L,         0x38 },
    { XK_Meta_R,        0xb8e0 },
    { XK_Alt_R,         0xb8e0 },

    { XK_Scroll_Lock,   0x46 },
    { XK_Caps_Lock,     0x3a }
};

static void x_refresh(DisplayState *ds)
{
  XEvent report;
  KeySym key;
  int mouse_update = 0;
  unsigned scan;
  XButtonEvent *button_event;
  XExposeEvent *expose_event;
  int count;
  static int gui_key_modifier_pressed;

  if (last_vm_running != vm_running) {
      last_vm_running = vm_running;
      x_SetCaption();
  }

  vga_update_display();

  while (XPending(qx_display) > 0)  {
    XNextEvent(qx_display, &report);

    switch  (report.type) {

    case Expose:
      expose_event = &report.xexpose;
      x_update(ds, expose_event->x, expose_event->y, expose_event->width, expose_event->height);
      break;

    case ButtonPress:
      button_event = (XButtonEvent *) &report;
      current_x = button_event->x;
      current_y = button_event->y;
      switch (button_event->button) {
	case Button1:
	  if(!gui_grab) {
		/* start grabbing all events */
		trigger_mouse_grab();
		break;
	  }
	  mouse_button_state |= 0x01;
	  mouse_update = 1;
	  break;
	case Button2:
	  mouse_button_state |= 0x04;
	  mouse_update = 1;
	  break;
	case Button3:
	  mouse_button_state |= 0x02;
	  mouse_update = 1;
	  break;
	}
      break;

    case ButtonRelease:
      button_event = (XButtonEvent *) &report;
      current_x = button_event->x;
      current_y = button_event->y;
      switch (button_event->button) {
	case Button1:
	  mouse_button_state &= ~0x01;
	  mouse_update = 1;
	  break;
	case Button2:
	  mouse_button_state &= ~0x04;
	  mouse_update = 1;
	  break;
	case Button3:
	  mouse_button_state &= ~0x02;
	  mouse_update = 1;
	  break;
	}
      break;

    case KeyPress:
    case KeyRelease:
      key = XLookupKeysym(&report.xkey, 0);

      if(key == XK_Shift_L) {
	if(report.type == KeyPress)
		gui_key_modifier_pressed |= 5;
	else
		gui_key_modifier_pressed &= ~1;
      } else if(key == XK_Control_L) {
	if(report.type == KeyPress)
		gui_key_modifier_pressed |= 6;
	else
		gui_key_modifier_pressed &= ~2;
      }
      if((gui_key_modifier_pressed & 7) == 7) {
	trigger_mouse_grab();
	gui_key_modifier_pressed &= ~4;
	break;
      }

      /* XK_F21 is for sun type 4 keyboards.. */
      if (key == XK_Break || key == XK_F21)
	  key = XK_Pause;

      if (key == XK_Pause) {
	  if (report.xkey.state & ControlMask)
	      scan = 0xc6e046e0;
	  else
	      scan = 0x451de1;
      }                   /* XK_F22 is sun type 4 PrtScr */
      else if (key == XK_Print || key == XK_F22) {
	  if (report.xkey.state & Mod1Mask)
	      scan = 0x54;
	  else
	      scan = 0x37e02ae0;
      }
#ifdef KBUK
      else if ((key == XK_numbersign) && (report.xkey.state & ShiftMask))
	  scan = 0x4;
#endif
      else if (key == XK_sterling)
	  scan = 0x4;
      else {
	  if (key >= 0x20 && key <= 0x20+sizeof(scan_table1))
		  scan = scan_table1[key - 0x20];
	   else   {
		  scan = 0;
		  for (count = 0; count <
		       (sizeof(scan_table2)/sizeof(scan_table2[0]));
		       count++)
			  if (scan_table2[count].key == key) {
				  scan = scan_table2[count].scan_code;
				  break;
				  }
		  }
	  if (scan == 0)
		  break;
	  }
      for (count = 0; scan; count++) {
	  if (report.type == KeyRelease)
	      kbd_put_keycode(0x80 | (scan & 0xff));
	  else
	      kbd_put_keycode(scan & 0xff);

	  scan >>= 8;
      }
      break;

    case MotionNotify:
      current_x = report.xmotion.x;
      current_y = report.xmotion.y;
      mouse_update = 1;
      break;

    default:
      break;
    } /* end switch */

    if (mouse_update && gui_grab) {
	send_keyboard_mouse_status();
	mouse_update = 0;
    }
  } /* end while */
}

void X_display_init(DisplayState *ds, int full_screen)
{
  XSizeHints size_hints;
  XWMHints wm_hints;
  XClassHint class_hints;
  XSetWindowAttributes attr;
  unsigned long valuemask = 0; /* ignore XGCvalues and use defaults */
  XGCValues values;
  char *progname = "qemu";
  XEvent report;

  if ( (qx_display=XOpenDisplay(NULL)) == NULL ) {
	fprintf(stderr, "Could not open X display\n");
	exit(1);
  }
  /* get screen size from display structure macro */
  q_x_screen_num = DefaultScreen(qx_display);

  /* create opaque window */
  win = XCreateSimpleWindow(qx_display, RootWindow(qx_display, q_x_screen_num),
    0, 0,     /* x y  would be settable from the command line or resource database. */
    720,      /* initial vga 80*25 9x16 */
    400,
    4,        /* border_width four pixels */
    BlackPixel(qx_display, q_x_screen_num),
    BlackPixel(qx_display, q_x_screen_num));

  default_depth  = DefaultDepth(qx_display, q_x_screen_num);
  default_visual = DefaultVisual(qx_display, q_x_screen_num);

  size_hints.flags =  PSize | PMinSize | PMaxSize;
  size_hints.max_width = size_hints.min_width = 720;
  size_hints.max_height = size_hints.min_height = 400;

  wm_hints.initial_state = NormalState;
  wm_hints.input = True;
  wm_hints.flags = StateHint | InputHint;

  class_hints.res_name = "qemu";
  class_hints.res_class = "Qemu";

  XSetWMProperties(qx_display, win,
		     NULL, NULL, // windowName iconName, set from x_SetCaption
		     &progname, 1, &size_hints, &wm_hints,
		     &class_hints);
  // (attempt to) enable backing store
  attr.save_under=1;
  attr.backing_store=Always;
  XChangeWindowAttributes(qx_display, win, CWSaveUnder|CWBackingStore, &attr);

  XSelectInput(qx_display, win,
		KeyPressMask | KeyReleaseMask |
		ButtonPressMask | ButtonReleaseMask | PointerMotionMask |
		ExposureMask| StructureNotifyMask);

  gc = XCreateGC(qx_display, win, valuemask, &values);

  XSetForeground(qx_display, gc, WhitePixel(qx_display, q_x_screen_num));
  XSetBackground(qx_display, gc, BlackPixel(qx_display, q_x_screen_num));

  /* Display window */
  XMapWindow(qx_display, win);
  XSync(qx_display, /* no discard */ 0);

  while (1) {
    XNextEvent(qx_display, &report);
    if (report.type == MapNotify) break;
  }
  ds->dpy_update = x_update;
  ds->dpy_resize = x_resize;
  ds->dpy_refresh = x_refresh;
  x_resize(ds, 720, 400);
//  trigger_mouse_grab();
}
